#!/usr/bin/env python
# -*- coding: utf-8 -*-
# _version.py
"""Package initialization for DVHA-Analytics."""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVH Analytics
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

__author__ = "Dan Cutright"
__email__ = "dan.cutright@gmail.com"
__version__ = "0.9.2.post2"
__release__ = "0.9.2.post2"
__version_info__ = (0, 9, 2)
